import os
import sys
import shutil
import json
import pandas as pd
import numpy as np
import logging
import datetime
import functools
import inspect
import string
import warnings

from pylg import TraceFunction
from pylg import trace

  
class ArgumentValidationError(ValueError):
    def __init__(self, arg_num, func_name, accepted_arg_type, given_arg_type, list_type):
        if(list_type):
            self.error = 'The {0} argument of {1}() is not in the list {2}, but is {3}'.format(arg_num,
                                                                         func_name,
                                                                         accepted_arg_type,
                                                                         given_arg_type)
        else:
            self.error = 'The {0} argument of {1}() is not a {2}, but is {3}'.format(arg_num,
                                                                     func_name,
                                                                     accepted_arg_type,
                                                                     given_arg_type)
 
    def __str__(self):
        return self.error
 
 
 
class InvalidArgumentNumberError(ValueError):
    def __init__(self, func_name):
        self.error = 'Invalid number of arguments for {0}()'.format(func_name)
 
    def __str__(self):
        return self.error
 
 
 
class InvalidReturnType(ValueError):
    def __init__(self, return_type, func_name):
        self.error = 'Invalid return type {0} for {1}()'.format(return_type,
                                                                func_name)
 
    def __str__(self):
        return self.error



def ordinal(num):
    if 10 <= num % 100 < 20:
        return '{0}th'.format(num)
    else:
        ord = {1 : 'st', 2 : 'nd', 3 : 'rd'}.get(num % 10, 'th')
        return '{0}{1}'.format(num, ord)


def accepts(*accepted_arg_types, **accepted_arg_dicts): 
    def accept_decorator(validate_function):
        @functools.wraps(validate_function)
        def decorator_wrapper(*function_args, **function_args_dicts):
            if len(accepted_arg_types) is not len(accepted_arg_types):
                raise InvalidArgumentNumberError(validate_function.__name__)
 
            # We're using enumerate to get the index, so we can pass the
            # argument number with the incorrect type to ArgumentValidationError.
            i = 0;
            for arg_num, (actual_arg, accepted_arg_type) in enumerate(zip(function_args, accepted_arg_types)):
                if(accepted_arg_type=="self"):
                    continue;
                if(type(accepted_arg_type)) == list:
                    if type(actual_arg) not in accepted_arg_type:
                        ord_num = ordinal(arg_num + 1)
                        if(accepted_arg_dicts["post_trace"]):
                            raise ArgumentValidationError(ord_num,
                                                      validate_function.function.function.__name__,
                                                      accepted_arg_type,
                                                      type(actual_arg),
                                                      True)
                        else:
                            raise ArgumentValidationError(ord_num,
                                                      validate_function.__name__,
                                                      accepted_arg_type,
                                                      type(actual_arg),
                                                      True)
                else:     
                    if not type(actual_arg) is accepted_arg_type:
                        ord_num = ordinal(arg_num + 1)
                        if(accepted_arg_dicts["post_trace"]):
                            raise ArgumentValidationError(ord_num,
                                                      validate_function.function.function.__name__,
                                                      accepted_arg_type,
                                                      type(actual_arg),
                                                      False)
                        else:
                            raise ArgumentValidationError(ord_num,
                                                      validate_function.__name__,
                                                      accepted_arg_type,
                                                      type(actual_arg),
                                                      False)
                i += 1;


            keys = list(function_args_dicts.keys());
            for i in range(len(keys)):
                func_arg_type = type(function_args_dicts[keys[i]]);
                accepted_arg_type = accepted_arg_dicts[keys[i]];
                if(type(accepted_arg_type) == list):
                    if(func_arg_type not in accepted_arg_type):
                        if(accepted_arg_dicts["post_trace"]):
                            raise ArgumentValidationError(keys[i],
                                                      validate_function.function.function.__name__,
                                                      accepted_arg_type,
                                                      func_arg_type,
                                                      True)
                        else:
                            raise ArgumentValidationError(keys[i],
                                                      validate_function.__name__,
                                                      accepted_arg_type,
                                                      func_arg_type,
                                                      True)
                else:
                    if(func_arg_type != accepted_arg_type):
                        if(accepted_arg_dicts["post_trace"]):
                            raise ArgumentValidationError(keys[i],
                                                      validate_function.function.function.__name__,
                                                      accepted_arg_type,
                                                      func_arg_type,
                                                      False)
                        else:
                            raise ArgumentValidationError(keys[i],
                                                      validate_function.__name__,
                                                      accepted_arg_type,
                                                      func_arg_type,
                                                      False)

            return validate_function(*function_args, **function_args_dicts)
        return decorator_wrapper
    return accept_decorator




class ConstraintError(ValueError):

    def __init__(self, msg):
        self.error = msg
 
    def __str__(self):
        return self.error



def ConstraintWarning(msg):
    warnings.warn(msg)




def error_checks(*arg_constraints, **kwargs_constraints):

    def check_gte(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float]):
            if(actual_value < limit):
                msg += "Value expected to be greater than equal to \"{}\", but is \"{}\"".format(limit, actual_value);
                raise ConstraintError(msg);
        if(type(actual_value) in [list, tuple]):
            for i in range(len(actual_value)):
                if(actual_value[i] < limit):
                    msg += "List's arg number \"{}\" expected to be greater than equal to \"{}\", but is \"{}\"".format(i+1, limit, actual_value[i]);
                    raise ConstraintError(msg);



    def check_gt(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float]):
            if(actual_value <= limit):
                msg += "Value expected to be strictly greater than to \"{}\", but is \"{}\"".format(limit, actual_value);
                raise ConstraintError(msg);
        if(type(actual_value) in [list, tuple]):
            for i in range(len(actual_value)):
                if(actual_value[i] <= limit):
                    msg += "List's arg number \"{}\" expected to be strictly greater than equal to \"{}\", but is \"{}\"".format(i+1, limit, actual_value[i]);
                    raise ConstraintError(msg);



    def check_lte(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float]):
            if(actual_value > limit):
                msg += "Value expected to be less than equal to \"{}\", but is \"{}\"".format(limit, actual_value);
                raise ConstraintError(msg);
        if(type(actual_value) in [list, tuple]):
            for i in range(len(actual_value)):
                if(actual_value[i] > limit):
                    msg += "List's arg number \"{}\" expected to be less than equal to \"{}\", but is \"{}\"".format(i+1, limit, actual_value[i]);
                    raise ConstraintError(msg);



    def check_lt(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float]):
            if(actual_value >= limit):
                msg += "Value expected to be strictly less than to \"{}\", but is \"{}\"".format(limit, actual_value);
                raise ConstraintError(msg);
        if(type(actual_value) in [list, tuple]):
            for i in range(len(actual_value)):
                if(actual_value[i] >= limit):
                    msg += "List's arg number \"{}\" expected to be strictly less than equal to \"{}\", but is \"{}\"".format(i+1, limit, actual_value[i]);
                    raise ConstraintError(msg);



    def check_eq(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float, str, list, tuple]):
            if(actual_value != limit):
                msg += "Value expected to be strictly equal to \"{}\", but is \"{}\"".format(limit, actual_value);
                raise ConstraintError(msg);

    def check_neq(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float, str, list, tuple]):
            if(actual_value == limit):
                msg += "Value expected to be strictly not equal to \"{}\", but is \"{}\"".format(limit, actual_value);
                raise ConstraintError(msg);

    def check_in(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in list(map(type, limit))):
            if(actual_value not in limit):
                msg += "Value expected to be one among \"{}\", but is \"{}\"".format(limit, actual_value);
                raise ConstraintError(msg);

    def check_nin(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in list(map(type, limit))):
            if(actual_value in limit):
                msg += "Value expected to be anything except \"{}\", but is \"{}\"".format(limit, actual_value);
                raise ConstraintError(msg);

    def check_folder(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) == str):
            if(not os.path.isdir(actual_value)):
                msg = "Folder \"{}\" not found".format(actual_value)
                raise ConstraintError(msg);
            if(limit == "r"):
                if(not os.access(actual_value, os.R_OK)):
                    msg = "Folder \"{}\" has no read access".format(actual_value)
                    raise ConstraintError(msg);
            if(limit == "w"):
                if(not os.access(actual_value, os.W_OK)):
                    msg = "Folder \"{}\" has no write access".format(actual_value)
                    raise ConstraintError(msg);
        if(type(actual_value) == list):
            for i in range(len(actual_value)):
                if(not os.path.isdir(actual_value[i])):
                    msg = "Folder \"{}\" not found".format(actual_value[i])
                    raise ConstraintError(msg);
                if(limit == "r"):
                    if(not os.access(actual_value[i], os.R_OK)):
                        msg = "Folder \"{}\" has no read access".format(actual_value[i])
                        raise ConstraintError(msg);
                if(limit == "w"):
                    if(not os.access(actual_value[i], os.W_OK)):
                        msg = "Folder \"{}\" has no write access".format(actual_value[i])
                        raise ConstraintError(msg);


    def check_file(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) == str):
            if(not os.path.isfile(actual_value)):
                msg = "File \"{}\" not found".format(actual_value)
                raise ConstraintError(msg);
        
            if(limit == "r"):
                if(not os.access(actual_value, os.R_OK)):
                    msg = "File \"{}\" has no read access".format(actual_value)
                    raise ConstraintError(msg);
            if(limit == "w"):
                if(not os.access(actual_value, os.W_OK)):
                    msg = "File \"{}\" has no write access".format(actual_value)
                    raise ConstraintError(msg);
        if(type(actual_value) == list):
            for i in range(len(actual_value)):
                if(not os.path.isfile(actual_value[i])):
                    msg = "File \"{}\" not found".format(actual_value[i])
                    raise ConstraintError(msg);
            
                if(limit == "r"):
                    if(not os.access(actual_value[i], os.R_OK)):
                        msg = "File \"{}\" has no read access".format(actual_value[i])
                        raise ConstraintError(msg);
                if(limit == "w"):
                    if(not os.access(actual_value[i], os.W_OK)):
                        msg = "File \"{}\" has no write access".format(actual_value[i])
                        raise ConstraintError(msg);


    def check_inc(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) == list):
            if(sorted(actual_value) != actual_value):
                msg += "List expected to be incremental, but is \"{}\"".format(actual_value);
                raise ConstraintError(msg);

    def check_dec(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) == list):
            if(sorted(actual_value, reverse=True) != actual_value):
                msg += "List expected to be decremental, but is \"{}\"".format(actual_value);
                raise ConstraintError(msg);


    def check_name(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) == str):
            total_list = [];
            for i in range(len(limit)):
                if(limit[i] == "a-z"):
                    total_list += list(string.ascii_lowercase)
                elif(limit[i] == "A-Z"):
                    total_list += list(string.ascii_uppercase)
                elif(limit[i] == "0-9"):
                    total_list += list(string.digits)
                else:
                    total_list += limit[i];
            
            actual_value = list(actual_value)
            for j in range(len(actual_value)):
                if(actual_value[j] not in total_list):
                    msg += "Character \"{}\" not allowed as per constrains \"{}\"".format(actual_value[j], limit);
                    raise ConstraintError(msg);



    def accept_decorator(validate_function):
        @functools.wraps(validate_function)
        def decorator_wrapper(*function_args, **function_args_dicts):
            if len(arg_constraints) is not len(function_args):
                raise InvalidArgumentNumberError(validate_function.__name__)

            if(kwargs_constraints["post_trace"]):
                function_name = validate_function.function.function.__name__;
            else:
                function_name = validate_function.__name__;

            for arg_num, (actual_arg, arg_constraint) in enumerate(zip(function_args, arg_constraints)):
                if(arg_constraint):
                    for i in range(len(arg_constraint)//2):
                        if(arg_constraint[i*2] == "gte"):
                            check_gte(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "gt"):
                            check_gt(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "lte"):
                            check_lte(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "lt"):
                            check_lt(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "eq"):
                            check_eq(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "neq"):
                            check_neq(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "in"):
                            check_in(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "nin"):
                            check_nin(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "folder"):
                            check_folder(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "file"):
                            check_file(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "inc"):
                            check_inc(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "dec"):
                            check_dec(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "name"):
                            check_name(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);

                                 


            keys = list(function_args_dicts.keys());
            for x in range(len(keys)):
                actual_arg = function_args_dicts[keys[x]];
                arg_constraint = kwargs_constraints[keys[x]];
                if(arg_constraint):
                    for i in range(len(arg_constraint)//2):
                        if(arg_constraint[i*2] == "gte"):
                            check_gte(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "gt"):
                            check_gt(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "lte"):
                            check_lte(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "lt"):
                            check_lt(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "eq"):
                            check_eq(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "neq"):
                            check_neq(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "in"):
                            check_in(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "nin"):
                            check_nin(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "folder"):
                            check_folder(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "file"):
                            check_file(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "inc"):
                            check_inc(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "dec"):
                            check_dec(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "name"):
                            check_name(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);



            return validate_function(*function_args, **function_args_dicts)
        return decorator_wrapper
    return accept_decorator



def warning_checks(*arg_constraints, **kwargs_constraints):

    def check_gte(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float]):
            if(actual_value < limit):
                msg += "Value expected to be greater than equal to \"{}\", but is \"{}\"".format(limit, actual_value);
                ConstraintWarning(msg);
        if(type(actual_value) in [list, tuple]):
            for i in range(len(actual_value)):
                if(actual_value[i] < limit):
                    msg += "List's arg number \"{}\" expected to be greater than equal to \"{}\", but is \"{}\"".format(i+1, limit, actual_value[i]);
                    ConstraintWarning(msg);


    def check_gt(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float]):
            if(actual_value <= limit):
                msg += "Value expected to be strictly greater than to \"{}\", but is \"{}\"".format(limit, actual_value);
                ConstraintWarning(msg);
        if(type(actual_value) in [list, tuple]):
            for i in range(len(actual_value)):
                if(actual_value[i] <= limit):
                    msg += "List's arg number \"{}\" expected to be strictly greater than equal to \"{}\", but is \"{}\"".format(i+1, limit, actual_value[i]);
                    ConstraintWarning(msg);



    def check_lte(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float]):
            if(actual_value > limit):
                msg += "Value expected to be less than equal to \"{}\", but is \"{}\"".format(limit, actual_value);
                ConstraintWarning(msg);
        if(type(actual_value) in [list, tuple]):
            for i in range(len(actual_value)):
                if(actual_value[i] > limit):
                    msg += "List's arg number \"{}\" expected to be less than equal to \"{}\", but is \"{}\"".format(i+1, limit, actual_value[i]);
                    ConstraintWarning(msg);



    def check_lt(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float]):
            if(actual_value >= limit):
                msg += "Value expected to be strictly less than to \"{}\", but is \"{}\"".format(limit, actual_value);
                ConstraintWarning(msg);
        if(type(actual_value) in [list, tuple]):
            for i in range(len(actual_value)):
                if(actual_value[i] >= limit):
                    msg += "List's arg number \"{}\" expected to be strictly less than equal to \"{}\", but is \"{}\"".format(i+1, limit, actual_value[i]);
                    ConstraintWarning(msg);


    def check_eq(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float, str, list, tuple]):
            if(actual_value != limit):
                msg += "Value expected to be strictly equal to \"{}\", but is \"{}\"".format(limit, actual_value);
                ConstraintWarning(msg);

    def check_neq(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in [int, float, str, list, tuple]):
            if(actual_value == limit):
                msg += "Value expected to be strictly not equal to \"{}\", but is \"{}\"".format(limit, actual_value);
                ConstraintWarning(msg);

    def check_in(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in list(map(type, limit))):
            if(actual_value not in limit):
                msg += "Value expected to be one among \"{}\", but is \"{}\"".format(limit, actual_value);
                ConstraintWarning(msg);

    def check_nin(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) in list(map(type, limit))):
            if(actual_value in limit):
                msg += "Value expected to be anything except \"{}\", but is \"{}\"".format(limit, actual_value);
                ConstraintWarning(msg);

    def check_folder(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) == str):
            if(not os.path.isdir(actual_value)):
                msg = "Folder \"{}\" not found".format(actual_value)
                ConstraintWarning(msg);
            if(limit == "r"):
                if(not os.access(actual_value, os.R_OK)):
                    msg = "Folder \"{}\" has no read access".format(actual_value)
                    ConstraintWarning(msg);
            if(limit == "w"):
                if(not os.access(actual_value, os.W_OK)):
                    msg = "Folder \"{}\" has no write access".format(actual_value)
                    ConstraintWarning(msg);
        if(type(actual_value) == list):
            for i in range(len(actual_value)):
                if(not os.path.isdir(actual_value[i])):
                    msg = "Folder \"{}\" not found".format(actual_value[i])
                    ConstraintWarning(msg);
                if(limit == "r"):
                    if(not os.access(actual_value[i], os.R_OK)):
                        msg = "Folder \"{}\" has no read access".format(actual_value[i])
                        ConstraintWarning(msg);
                if(limit == "w"):
                    if(not os.access(actual_value[i], os.W_OK)):
                        msg = "Folder \"{}\" has no write access".format(actual_value[i])
                        ConstraintWarning(msg);


    def check_file(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) == str):
            if(not os.path.isdir(actual_value)):
                msg = "File \"{}\" not found".format(actual_value)
                ConstraintWarning(msg);
            if(limit == "r"):
                if(not os.access(actual_value, os.R_OK)):
                    msg = "File \"{}\" has no read access".format(actual_value)
                    ConstraintWarning(msg);
            if(limit == "w"):
                if(not os.access(actual_value, os.W_OK)):
                    msg = "File \"{}\" has no write access".format(actual_value)
                    ConstraintWarning(msg);
        if(type(actual_value) == list):
            for i in range(len(actual_value)):
                if(not os.path.isdir(actual_value[i])):
                    msg = "File \"{}\" not found".format(actual_value[i])
                    ConstraintWarning(msg);
                if(limit == "r"):
                    if(not os.access(actual_value[i], os.R_OK)):
                        msg = "File \"{}\" has no read access".format(actual_value[i])
                        ConstraintWarning(msg);
                if(limit == "w"):
                    if(not os.access(actual_value[i], os.W_OK)):
                        msg = "File \"{}\" has no write access".format(actual_value[i])
                        ConstraintWarning(msg);


    def check_inc(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) == list):
            if(sorted(actual_value) != actual_value):
                msg += "List expected to be incremental, but is \"{}\"".format(actual_value);
                ConstraintWarning(msg);

    def check_dec(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) == list):
            if(sorted(actual_value, reverse=True) != actual_value):
                msg += "List expected to be decremental, but is \"{}\"".format(actual_value);
                ConstraintWarning(msg);


    def check_name(actual_value, limit, function_name, arg_num=None, arg_name=None):
        if(arg_num):
            arg = arg_num;
            msg = "Constraint Mismatch for argument number \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(arg_name):
            arg = arg_name;
            msg = "Constraint Mismatch for argument name \"{}\" in function \"{}\".\n".format(arg, function_name);
        if(type(actual_value) == str):
            total_list = [];
            for i in range(len(limit)):
                if(limit[i] == "a-z"):
                    total_list += list(string.ascii_lowercase)
                elif(limit[i] == "A-Z"):
                    total_list += list(string.ascii_uppercase)
                elif(limit[i] == "0-9"):
                    total_list += list(string.digits)
                else:
                    total_list += limit[i];
            
            actual_value = list(actual_value)
            for j in range(len(actual_value)):
                if(actual_value[j] not in total_list):
                    msg += "Character \"{}\" not allowed as per constrains \"{}\"".format(actual_value[j], limit);
                    ConstraintWarning(msg);



    def accept_decorator(validate_function):
        @functools.wraps(validate_function)
        def decorator_wrapper(*function_args, **function_args_dicts):
            if len(arg_constraints) is not len(function_args):
                raise InvalidArgumentNumberError(validate_function.__name__)

            if(kwargs_constraints["post_trace"]):
                function_name = validate_function.function.function.__name__;
            else:
                function_name = validate_function.__name__;

            for arg_num, (actual_arg, arg_constraint) in enumerate(zip(function_args, arg_constraints)):
                if(arg_constraint):
                    for i in range(len(arg_constraint)//2):
                        if(arg_constraint[i*2] == "gte"):
                            check_gte(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "gt"):
                            check_gt(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "lte"):
                            check_lte(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "lt"):
                            check_lt(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "eq"):
                            check_eq(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "neq"):
                            check_neq(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "in"):
                            check_in(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "nin"):
                            check_nin(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "folder"):
                            check_folder(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "file"):
                            check_file(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "inc"):
                            check_inc(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "dec"):
                            check_dec(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);
                        if(arg_constraint[i*2] == "name"):
                            check_name(actual_arg, arg_constraint[i*2+1], function_name, arg_num=arg_num+1);

                                 


            keys = list(function_args_dicts.keys());
            for x in range(len(keys)):
                actual_arg = function_args_dicts[keys[x]];
                arg_constraint = kwargs_constraints[keys[x]];
                if(arg_constraint):
                    for i in range(len(arg_constraint)//2):
                        if(arg_constraint[i*2] == "gte"):
                            check_gte(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "gt"):
                            check_gt(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "lte"):
                            check_lte(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "lt"):
                            check_lt(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "eq"):
                            check_eq(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "neq"):
                            check_neq(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "in"):
                            check_in(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "nin"):
                            check_nin(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "folder"):
                            check_folder(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "file"):
                            check_file(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "inc"):
                            check_inc(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "dec"):
                            check_dec(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);
                        if(arg_constraint[i*2] == "name"):
                            check_name(actual_arg, arg_constraint[i*2+1], function_name, arg_name=keys[x]);



            return validate_function(*function_args, **function_args_dicts)
        return decorator_wrapper
    return accept_decorator




from monk.tf_keras_1.finetune.imports import *
from monk.system.imports import *

from monk.tf_keras_1.finetune.level_10_schedulers_main import prototype_schedulers


class prototype_optimizers(prototype_schedulers):
    '''
    Main class for all optimizers in expert mode

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);


    ###############################################################################################################################################
    @warning_checks(None, ["lt", 1], momentum=["lt", 1.5], weight_decay=["lt", 0.01], momentum_dampening_rate=None, 
        clipnorm=None, clipvalue=None, post_trace=False)
    @error_checks(None, ["gt", 0], momentum=["gte", 0], weight_decay=["gte", 0], momentum_dampening_rate=None, 
        clipnorm=None, clipvalue=None, post_trace=False)
    @accepts("self", [int, float], momentum=[int, float], weight_decay=[int, float], momentum_dampening_rate=[int, float], 
        clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def optimizer_sgd(self, learning_rate, momentum=0, weight_decay=0, momentum_dampening_rate=0, clipnorm=0.0, clipvalue=0.0):
        '''
        Select stochastic gradient descent optimizer

        Args:
            learning_rate (float): Initial base learning rate
            momentum (float): Momentum value for driving the weights towards minima
            weight_decay (float): Value for regularizing weights post every update
            momentum_dampening_rate (float): Reduction rate for momentum
            clipnorm (float): Gradient clipping factor
            clipvalue (float): Value for clipping

        Returns:
            None
        '''
        self.system_dict = sgd(self.system_dict, learning_rate, 
                momentum=momentum, weight_decay=weight_decay, momentum_dampening_rate=momentum_dampening_rate, clipnorm=clipnorm, clipvalue=clipvalue);
        self.custom_print("Optimizer");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["optimizer"]["name"]));
        self.custom_print("    Learning rate: {}".format(self.system_dict["hyper-parameters"]["learning_rate"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["optimizer"]["params"]));
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: clipnorm and clipvalue are active only for keras in current version of Monk");
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: momentum_dampening_rate is active only for pytorch in current version of Monk");
        self.custom_print("");
    ###############################################################################################################################################


    ###############################################################################################################################################
    @warning_checks(None, ["lt", 1], momentum=["lt", 1.5], weight_decay=["lt", 0.01], momentum_dampening_rate=None, 
        clipnorm=None, clipvalue=None, post_trace=False)
    @error_checks(None, ["gt", 0], momentum=["gte", 0], weight_decay=["gte", 0], momentum_dampening_rate=None, 
        clipnorm=None, clipvalue=None, post_trace=False)
    @accepts("self", [int, float], momentum=[int, float], weight_decay=[int, float], momentum_dampening_rate=[int, float], 
        clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def optimizer_nesterov_sgd(self, learning_rate, momentum=0, weight_decay=0, momentum_dampening_rate=0, clipnorm=0.0, clipvalue=0.0):
        '''
        Select stochastic gradient descent optimizer with nesterov acceleration

        Args:
            learning_rate (float): Initial base learning rate
            momentum (float): Momentum value for driving the weights towards minima
            weight_decay (float): Value for regularizing weights post every update
            momentum_dampening_rate (float): Reduction rate for momentum
            clipnorm (float): Gradient clipping factor
            clipvalue (float): Value for clipping

        Returns:
            None
        '''
        self.system_dict = nesterov_sgd(self.system_dict, learning_rate, 
                momentum=momentum, weight_decay=weight_decay, momentum_dampening_rate=momentum_dampening_rate, clipnorm=clipnorm, clipvalue=clipvalue);
        self.custom_print("Optimizer");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["optimizer"]["name"]));
        self.custom_print("    Learning rate: {}".format(self.system_dict["hyper-parameters"]["learning_rate"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["optimizer"]["params"]));
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: clipnorm and clipvalue are active only for keras in current version of Monk");
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: momentum_dampening_rate is active only for pytorch in current version of Monk");
        self.custom_print("");
    ###############################################################################################################################################



    ###############################################################################################################################################
    @warning_checks(None, ["lt", 1], decay_rate=["lt", 1], epsilon=["lt", 0.001], weight_decay=["lt", 0.01], 
        clipnorm=None, clipvalue=None, post_trace=None)
    @error_checks(None, ["gt", 0], decay_rate=["gt", 0], epsilon=["gte", 0], weight_decay=["gte", 0], 
        clipnorm=None, clipvalue=None, post_trace=False)
    @accepts("self", [int, float], decay_rate=[int, float], epsilon=[int, float], weight_decay=[int, float], 
        clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def optimizer_rmsprop(self, learning_rate, decay_rate=0.99, epsilon=1e-08, weight_decay=0, clipnorm=0.0, clipvalue=0.0):
        '''
        Select root mean score prop optimizer

        Args:
            learning_rate (float): Initial base learning rate
            decay_rate (float): A decay factor of moving average over past squared gradient.
            epsilon (float): A value to avoid division by zero
            weight_decay (float): Value for regularizing weights post every update
            clipnorm (float): Gradient clipping factor
            clipvalue (float): Value for clipping

        Returns:
            None
        '''
        self.system_dict = rmsprop(self.system_dict , learning_rate, 
            decay_rate=decay_rate, epsilon=epsilon, weight_decay=weight_decay, clipnorm=clipnorm, clipvalue=clipvalue);
        self.custom_print("Optimizer");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["optimizer"]["name"]));
        self.custom_print("    Learning rate: {}".format(self.system_dict["hyper-parameters"]["learning_rate"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["optimizer"]["params"]));
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: clipnorm and clipvalue are active only for keras in current version of Monk");
        self.custom_print("");
    ###############################################################################################################################################



    ###############################################################################################################################################
    @warning_checks(None, ["lt, 1"], beta1=["lt", 1], beta2=["lt", 1], epsilon=["lt", 0.001],  weight_decay=["lt", 0.01], amsgrad=None, 
        clipnorm=None, clipvalue=None, post_trace=False)
    @error_checks(None, ["gt", 0], beta1=["gte", 0], beta2=["gte", 0], epssilon=["gte", 0], weight_decay=["gte", 0], amsgrad=None, 
        clipnorm=None, clipvalue=None, post_trace=False)
    @accepts("self", [int, float], beta1=[int, float], beta2=[int, float], epsilon=[int, float], weight_decay=[int, float], amsgrad=bool, 
        clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def optimizer_adam(self, learning_rate, beta1=0.9, beta2=0.999, epsilon=1e-08, weight_decay=0, amsgrad=False, 
        clipnorm=0.0, clipvalue=0.0):
        '''
        Select ADAM optimizer

        Args:
            learning_rate (float): Initial base learning rate
            beta1 (float): Exponential decay rate for first momentum estimates
            beta2 (float): Exponential decay rate for first second estimates
            weight_decay (float): Value for regularizing weights post every update
            amsgrad (bool): If True, AMSGrad variant of this algorithm is used
            epsilon (float): A value to avoid division by zero
            clipnorm (float): Gradient clipping factor
            clipvalue (float): Value for clipping

        Returns:
            None
        '''
        self.system_dict = adam(self.system_dict, learning_rate,
            beta1=beta1, beta2=beta2, epsilon=epsilon, weight_decay=weight_decay, amsgrad=amsgrad, clipnorm=clipnorm, clipvalue=clipvalue);
        
        self.custom_print("Optimizer");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["optimizer"]["name"]));
        self.custom_print("    Learning rate: {}".format(self.system_dict["hyper-parameters"]["learning_rate"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["optimizer"]["params"]));
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: clipnorm and clipvalue are active only for keras in current version of Monk");
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: amsgrad is active only for keras and pytorch in current version of Monk");
        self.custom_print("");
    ###############################################################################################################################################


    ###############################################################################################################################################
    @warning_checks(None, ["lt, 1"], beta1=["lt", 1], beta2=["lt", 1], epsilon=["lt", 0.001],  weight_decay=["lt", 0.01], amsgrad=None, 
        momentum_decay=None, clipnorm=None, clipvalue=None, post_trace=False)
    @error_checks(None, ["gt", 0], beta1=["gte", 0], beta2=["gte", 0], epssilon=["gte", 0], weight_decay=["gte", 0], amsgrad=None, 
        momentum_decay=None, clipnorm=None, clipvalue=None, post_trace=False)
    @accepts("self", [int, float], beta1=[int, float], beta2=[int, float], epsilon=[int, float], weight_decay=[int, float], amsgrad=bool, 
        momentum_decay=[int, float], clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def optimizer_nesterov_adam(self, learning_rate, beta1=0.9, beta2=0.999, epsilon=1e-08, weight_decay=0, amsgrad=False,
        momentum_decay=0.004, clipnorm=0.0, clipvalue=0.0):
        '''
        Select ADAM optimizer with nesterov momentum acceleration

        Args:
            learning_rate (float): Initial base learning rate
            beta1 (float): Exponential decay rate for first momentum estimates
            beta2 (float): Exponential decay rate for first second estimates
            weight_decay (float): Value for regularizing weights post every update
            amsgrad (bool): If True, AMSGrad variant of this algorithm is used
            epsilon (float): A value to avoid division by zero
            clipnorm (float): Gradient clipping factor
            clipvalue (float): Value for clipping

        Returns:
            None
        '''
        self.system_dict = nesterov_adam(self.system_dict, learning_rate,
            beta1=beta1, beta2=beta2, epsilon=epsilon, weight_decay=weight_decay, amsgrad=amsgrad,
            momentum_decay=momentum_decay, clipnorm=clipnorm, clipvalue=clipvalue);
        
        self.custom_print("Optimizer");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["optimizer"]["name"]));
        self.custom_print("    Learning rate: {}".format(self.system_dict["hyper-parameters"]["learning_rate"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["optimizer"]["params"]));
        self.custom_print("");
        ConstraintWarning("OptimizerWarning: nesterov adam is active only for keras and gluon in current version of Monk");
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: amsgrad is inactive in current version of Monk");
        self.custom_print("");
    ###############################################################################################################################################


    ###############################################################################################################################################
    @warning_checks(None, ["lt", 1], beta1=["lt", 1], beta2=["lt", 1], epsilon=["lt", 0.001], weight_decay=["lt", 0.01], 
        clipnorm=None, clipvalue=None, post_trace=False)
    @error_checks(None, ["gt", 0], beta1=["gte", 0], beta2=["gte", 0], epsilon=["gte", 0], weight_decay=["gte", 0], 
        clipnorm=None, clipvalue=None, post_trace=False)
    @accepts("self", [int, float], beta1=[int, float], beta2=[int, float], epsilon=[int, float], weight_decay=[int, float], 
        clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def optimizer_adamax(self, learning_rate, beta1=0.9, beta2=0.999, epsilon=1e-08, weight_decay=0,
        clipnorm=0.0, clipvalue=0.0):
        '''
        Select Adamax optimizer

        Args:
            learning_rate (float): Initial base learning rate
            beta1 (float): Exponential decay rate for first momentum estimates
            beta2 (float): Exponential decay rate for first second estimates
            weight_decay (float): Value for regularizing weights post every update
            epsilon (float): A value to avoid division by zero
            clipnorm (float): Gradient clipping factor
            clipvalue (float): Value for clipping

        Returns:
            None
        '''

        self.system_dict = adamax(self.system_dict, learning_rate,
            beta1=beta1, beta2=beta2, epsilon=epsilon, weight_decay=weight_decay, clipnorm=clipnorm, clipvalue=clipvalue);  
        
        self.custom_print("Optimizer");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["optimizer"]["name"]));
        self.custom_print("    Learning rate: {}".format(self.system_dict["hyper-parameters"]["learning_rate"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["optimizer"]["params"]));
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: clipnorm and clipvalue are active only for keras in current version of Monk");
        self.custom_print("");
    ###############################################################################################################################################


    ###############################################################################################################################################
    @warning_checks(None, ["lt", 1], rho=["lt", 1], epsilon=["lt", 0.001], weight_decay=["lt", 0.01], 
        clipnorm=None, clipvalue=None, post_trace=False)
    @error_checks(None, ["gt", 0], rho=["gt", 0], epsilon=["gte", 0], weight_decay=["gte", 0], 
        clipnorm=None, clipvalue=None, post_trace=False)
    @accepts("self", [int, float], rho=[int, float], epsilon=[int, float], weight_decay=[int, float], 
        clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def optimizer_adadelta(self, learning_rate, rho=0.9, epsilon=1e-06, weight_decay=0, 
        clipnorm=0.0, clipvalue=0.0):
        '''
        Select Adadelta optimizer

        Args:
            learning_rate (float): Initial base learning rate
            rho (float): Exponential decay rate for momentum estimates
            weight_decay (float): Value for regularizing weights post every update
            epsilon (float): A value to avoid division by zero
            clipnorm (float): Gradient clipping factor
            clipvalue (float): Value for clipping

        Returns:
            None
        '''
        self.system_dict = adadelta(self.system_dict, learning_rate,
            rho=rho, epsilon=epsilon, weight_decay=weight_decay, clipnorm=clipnorm, clipvalue=clipvalue);
        
        self.custom_print("Optimizer");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["optimizer"]["name"]));
        self.custom_print("    Learning rate: {}".format(self.system_dict["hyper-parameters"]["learning_rate"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["optimizer"]["params"]));
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: clipnorm and clipvalue are active only for keras in current version of Monk");
        self.custom_print("");
    ###############################################################################################################################################


    ###############################################################################################################################################
    @warning_checks(None, ["lt", 1], learning_rate_decay=None, weight_decay=["lt", 0.01], epsilon=None, 
        clipnorm=None, clipvalue=None, post_trace=False)
    @error_checks(None, ["gt", 0], learning_rate_decay=None, weight_decay=["gte", 0], epsilon=None, 
        clipnorm=None, clipvalue=None, post_trace=False)
    @accepts("self", [int, float], learning_rate_decay=[int, float], weight_decay=[int, float], 
        epsilon=[int, float], clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def optimizer_adagrad(self, learning_rate, learning_rate_decay=0, weight_decay=0, epsilon=1e-08,
        clipnorm=0.0, clipvalue=0.0):
        '''
        Select Adagrad optimizer

        Args:
            learning_rate (float): Initial base learning rate
            learning_rate_decay (float): Learning rate decay factor
            weight_decay (float): Value for regularizing weights post every update
            epsilon (float): A value to avoid division by zero
            clipnorm (float): Gradient clipping factor
            clipvalue (float): Value for clipping

        Returns:
            None
        '''
        self.system_dict = adagrad(self.system_dict, learning_rate,
            learning_rate_decay=learning_rate_decay, weight_decay=weight_decay, epsilon=epsilon, 
            clipnorm=clipnorm, clipvalue=clipvalue);
        
        self.custom_print("Optimizer");
        self.custom_print("    Name:          {}".format(self.system_dict["hyper-parameters"]["optimizer"]["name"]));
        self.custom_print("    Learning rate: {}".format(self.system_dict["hyper-parameters"]["learning_rate"]));
        self.custom_print("    Params:        {}".format(self.system_dict["hyper-parameters"]["optimizer"]["params"]));
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: clipnorm and clipvalue are active only for keras in current version of Monk");
        self.custom_print("");
        ConstraintWarning("ArgumentWarning: learning_rate_decay is active only for pytorch in current version of Monk");
        self.custom_print("");
    ###############################################################################################################################################



























import os
import sys

import psutil

from monk.gluon_prototype import prototype
from monk.compare_prototype import compare
from monk.pip_unit_tests.gluon.common import print_start
from monk.pip_unit_tests.gluon.common import print_status


def test_optimizer_rmsprop(system_dict):
    forward = True;
    if(not os.path.isdir("datasets")):
        os.system("! wget --load-cookies /tmp/cookies.txt \"https://docs.google.com/uc?export=download&confirm=$(wget --save-cookies /tmp/cookies.txt --keep-session-cookies --no-check-certificate 'https://docs.google.com/uc?export=download&id=1rG-U1mS8hDU7_wM56a1kc-li_zHLtbq2' -O- | sed -rn 's/.*confirm=([0-9A-Za-z_]+).*/\1\n/p')&id=1rG-U1mS8hDU7_wM56a1kc-li_zHLtbq2\" -O datasets.zip && rm -rf /tmp/cookies.txt")
        os.system("! unzip -qq datasets.zip")

    test = "test_optimizer_rmsprop";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            gtf = prototype(verbose=0);
            gtf.Prototype("sample-project-1", "sample-experiment-1");
            gtf.Default(dataset_path="datasets/dataset_cats_dogs_train", 
                model_name="resnet18_v1", freeze_base_network=True, num_epochs=2);
            gtf.optimizer_rmsprop(0.01, weight_decay=0.0001, decay_rate=0.9, 
            	clipnorm=1.0, clipvalue=0.5);
            gtf.Train();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");

    return system_dict



from monk.system.imports import *
from monk.gluon.finetune.imports import *

from monk.system.base_class import system


class finetune_dataset(system):
    '''
    Base class for dataset params

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''

    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);

    ###############################################################################################################################################
    @accepts("self", test=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_dataset_dataloader(self, test=False):
        '''
        Setup the dataloader.

        Args:
            test (bool): If True then test data is loaded, else train data is loaded.

        Returns:
            None
        '''
        if(test):
            label_type = self.system_dict["dataset"]["label_type"];
            
            if(label_type == "single" or label_type == False):
                num_workers = self.system_dict["dataset"]["params"]["num_workers"];
                if(self.system_dict["dataset"]["params"]["dataset_test_type"] == "foldered"):
                    test_dataset = mx.gluon.data.vision.ImageFolderDataset(self.system_dict["dataset"]["test_path"]).transform_first(self.system_dict["local"]["data_transforms"]["test"]);
                    if(not self.system_dict["dataset"]["params"]["classes"]):
                        self.system_dict["dataset"]["params"]["classes"] = list(np.unique(sorted(os.listdir(self.system_dict["dataset"]["test_path"]))));
                elif(self.system_dict["dataset"]["params"]["dataset_test_type"] == "csv"):
                    if(not self.system_dict["dataset"]["params"]["classes"]):
                        img_list, label_list, self.system_dict["dataset"]["params"]["classes"] = parse_csv(self.system_dict["dataset"]["csv_test"], self.system_dict["dataset"]["params"]["test_delimiter"]);
                    else:
                        img_list, label_list, _ = parse_csv(self.system_dict["dataset"]["csv_test"], self.system_dict["dataset"]["params"]["test_delimiter"]);
                    test_dataset = DatasetCustom(img_list, label_list, self.system_dict["dataset"]["test_path"]).transform_first(self.system_dict["local"]["data_transforms"]["test"]);
                self.system_dict["local"]["data_loaders"]['test'] = mx.gluon.data.DataLoader(test_dataset, batch_size=1, shuffle = False, num_workers=num_workers);
                self.system_dict["dataset"]["params"]["num_test_images"] = len(self.system_dict["local"]["data_loaders"]['test']);

            else:
                num_workers = self.system_dict["dataset"]["params"]["num_workers"];
                if(not self.system_dict["dataset"]["params"]["classes"]):
                    img_list, label_list, self.system_dict["dataset"]["params"]["classes"] = parse_csv_updated(self.system_dict["dataset"]["csv_test"], 
                                                                                                    self.system_dict["dataset"]["params"]["test_delimiter"]);
                else:
                    img_list, label_list, _ = parse_csv_updated(self.system_dict["dataset"]["csv_test"], self.system_dict["dataset"]["params"]["test_delimiter"]);

                test_dataset = DatasetCustomMultiLabel(img_list, label_list, 
                                                        self.system_dict["dataset"]["params"]["classes"], 
                                                        self.system_dict["dataset"]["test_path"]).transform_first(self.system_dict["local"]["data_transforms"]["test"]);

                self.system_dict["local"]["data_loaders"]['test'] = mx.gluon.data.DataLoader(test_dataset, batch_size=1, shuffle = False, num_workers=num_workers);
                self.system_dict["dataset"]["params"]["num_test_images"] = len(self.system_dict["local"]["data_loaders"]['test']);

        else:
            sampled_dataset = None;
            image_datasets = {};
            dataset_type = self.system_dict["dataset"]["dataset_type"];
            label_type = self.system_dict["dataset"]["label_type"];
            dataset_train_path = self.system_dict["dataset"]["train_path"];
            dataset_val_path = self.system_dict["dataset"]["val_path"];
            csv_train = self.system_dict["dataset"]["csv_train"];
            csv_val = self.system_dict["dataset"]["csv_val"];
            train_val_split = self.system_dict["dataset"]["params"]["train_val_split"];
            delimiter = self.system_dict["dataset"]["params"]["delimiter"];
            batch_size = self.system_dict["dataset"]["params"]["batch_size"];
            shuffle = self.system_dict["dataset"]["params"]["train_shuffle"];
            num_workers = self.system_dict["dataset"]["params"]["num_workers"];



            if(label_type == "single" or label_type == False):
                if(dataset_type == "train"):
                    sampled_dataset = mx.gluon.data.vision.ImageFolderDataset(dataset_train_path).transform_first(self.system_dict["local"]["data_transforms"]["train"]);
                elif(dataset_type == "train-val"):
                    image_datasets["train"] = mx.gluon.data.vision.ImageFolderDataset(dataset_train_path).transform_first(self.system_dict["local"]["data_transforms"]["train"]);
                    image_datasets["val"] = mx.gluon.data.vision.ImageFolderDataset(dataset_val_path).transform_first(self.system_dict["local"]["data_transforms"]["val"]);
                elif(dataset_type == "csv_train"):
                    img_list, label_list, self.system_dict["dataset"]["params"]["classes"] = parse_csv(csv_train, delimiter);
                    sampled_dataset = DatasetCustom(img_list, label_list, dataset_train_path).transform_first(self.system_dict["local"]["data_transforms"]["train"]);
                elif(dataset_type == "csv_train-val"):
                    img_list, label_list, self.system_dict["dataset"]["params"]["classes"] = parse_csv(csv_train, delimiter);
                    image_datasets["train"] = DatasetCustom(img_list, label_list, dataset_train_path).transform_first(self.system_dict["local"]["data_transforms"]["train"]);
                    img_list, label_list, self.system_dict["dataset"]["params"]["classes"] = parse_csv(csv_val, delimiter);
                    image_datasets["val"] = DatasetCustom(img_list, label_list, dataset_val_path).transform_first(self.system_dict["local"]["data_transforms"]["val"]);



                if(sampled_dataset):
                    lengths = [int(len(sampled_dataset)*train_val_split), len(sampled_dataset) - int(len(sampled_dataset)*train_val_split)]
                    image_datasets["train"], image_datasets["val"] = torch.utils.data.dataset.random_split(sampled_dataset, lengths)


                if("csv" not in dataset_type):
                    self.system_dict["dataset"]["params"]["classes"] = list(np.unique(sorted(os.listdir(dataset_train_path))));

                self.system_dict["dataset"]["params"]["num_classes"] = len(self.system_dict["dataset"]["params"]["classes"]);

                
                if(self.system_dict["dataset"]["params"]["weighted_sample"]):
                    self.custom_print("    Weighted sampling enabled.");
                    self.custom_print("    Weighted sampling temporarily suspended. Skipping step");


                self.system_dict["dataset"]["params"]["num_train_images"] = len(image_datasets["train"]);
                self.system_dict["dataset"]["params"]["num_val_images"] = len(image_datasets["val"]);


                if(self.system_dict["dataset"]["params"]["weighted_sample"]):
                    self.system_dict["local"]["data_loaders"]["train"] = mx.gluon.data.DataLoader(image_datasets["train"],
                                                                    batch_size=batch_size, shuffle=True, num_workers=num_workers);
                else:
                    self.system_dict["local"]["data_loaders"]["train"] = mx.gluon.data.DataLoader(image_datasets["train"],
                                                                    batch_size=batch_size, shuffle=shuffle, num_workers=num_workers);

                self.system_dict["local"]["data_loaders"]["val"] = mx.gluon.data.DataLoader(image_datasets["val"],
                                                                    batch_size=batch_size, shuffle=shuffle, num_workers=num_workers);

                

                self.system_dict["dataset"]["status"]= True;

            else:

                if(dataset_type == "csv_train"):
                    img_list, label_list, self.system_dict["dataset"]["params"]["classes"] = parse_csv_updated(csv_train, delimiter);
                    sampled_dataset = DatasetCustomMultiLabel(img_list, label_list, 
                                                                self.system_dict["dataset"]["params"]["classes"], 
                                                                dataset_train_path).transform_first(self.system_dict["local"]["data_transforms"]["train"]);
                elif(dataset_type == "csv_train-val"):
                    img_list, label_list, self.system_dict["dataset"]["params"]["classes"] = parse_csv_updated(csv_train, delimiter);
                    image_datasets["train"] = DatasetCustomMultiLabel(img_list, label_list, 
                                                                self.system_dict["dataset"]["params"]["classes"], 
                                                                dataset_train_path).transform_first(self.system_dict["local"]["data_transforms"]["train"]);
                    img_list, label_list, _ = parse_csv_updated(csv_val, delimiter);
                    image_datasets["val"] = DDatasetCustomMultiLabel(img_list, label_list, 
                                                                self.system_dict["dataset"]["params"]["classes"], 
                                                                dataset_val_path).transform_first(self.system_dict["local"]["data_transforms"]["val"]);

                if(sampled_dataset):
                    lengths = [int(len(sampled_dataset)*train_val_split), len(sampled_dataset) - int(len(sampled_dataset)*train_val_split)]
                    image_datasets["train"], image_datasets["val"] = torch.utils.data.dataset.random_split(sampled_dataset, lengths)

                self.system_dict["dataset"]["params"]["num_classes"] = len(self.system_dict["dataset"]["params"]["classes"]);

                self.system_dict["dataset"]["params"]["num_train_images"] = len(image_datasets["train"]);
                self.system_dict["dataset"]["params"]["num_val_images"] = len(image_datasets["val"]);


                if(self.system_dict["dataset"]["params"]["weighted_sample"]):
                    self.system_dict["local"]["data_loaders"]["train"] = mx.gluon.data.DataLoader(image_datasets["train"],
                                                                    batch_size=batch_size, shuffle=True, num_workers=num_workers);
                else:
                    self.system_dict["local"]["data_loaders"]["train"] = mx.gluon.data.DataLoader(image_datasets["train"],
                                                                    batch_size=batch_size, shuffle=shuffle, num_workers=num_workers);

                self.system_dict["local"]["data_loaders"]["val"] = mx.gluon.data.DataLoader(image_datasets["val"],
                                                                    batch_size=batch_size, shuffle=shuffle, num_workers=num_workers);

                

                self.system_dict["dataset"]["status"]= True;



    ###############################################################################################################################################


    ###############################################################################################################################################
    @accepts("self", test=bool, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_dataset_final(self, test=False):
        '''
        Set the transforms and then invoke data loader.

        Args:
            test (bool): If True then test tranforms are set and test dataloader is prepared data, 
                        else train transforms are set and train dataloader is prepared.

        Returns:
            None
        '''
        if(test):
            self.system_dict = set_transform_test(self.system_dict);
        else:
            self.system_dict = set_transform_trainval(self.system_dict);
        self.set_dataset_dataloader(test=test);
    ###############################################################################################################################################


from monk.pytorch.schedulers.imports import *
from monk.system.imports import *

@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def retrieve_scheduler(system_dict):
    '''
    Retrieve schedulers in validation, testing, copy-from and resume states

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    if(system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"] == "fixed"):
        system_dict["local"]["learning_rate_scheduler"] = None;
    else:
        system_dict["local"]["learning_rate_scheduler"] = system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"];

    return system_dict;


from monk.gluon.schedulers.imports import *
from monk.system.imports import *


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def load_scheduler(system_dict):
    '''
    Load schedulers for training state

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    learning_rate_scheduler = system_dict["local"]["learning_rate_scheduler"];
    num_batches = len(system_dict["local"]["data_loaders"]["train"]);
    learning_rate = system_dict["hyper-parameters"]["learning_rate"];
    if(learning_rate_scheduler == "steplr"):
        step_size = system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["step_size"]
        if(step_size > system_dict["hyper-parameters"]["num_epochs"]):
            msg = "Step size - {} > Num epochs - {}\n".format(step_size, system_dict["hyper-parameters"]["num_epochs"]);
            msg += "Change scheduler step size";
            raise ConstraintError(msg);

        system_dict["local"]["learning_rate_scheduler"] = mx.lr_scheduler.FactorScheduler(
            step = system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["step_size"]*num_batches, 
            factor=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["gamma"], 
            base_lr=learning_rate, 
            warmup_steps=0, warmup_begin_lr=0, warmup_mode='linear', stop_factor_lr=1e-08);

    elif(learning_rate_scheduler == "multisteplr"):
        max_step_size = max(system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["milestones"])
        if(max_step_size > system_dict["hyper-parameters"]["num_epochs"]):
            msg = "Step size - {} > Num epochs - {}\n".format(max_step_size, system_dict["hyper-parameters"]["num_epochs"]);
            msg += "Change scheduler step size";
            raise ConstraintError(msg);

        steps = system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["milestones"];
        milestones = system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["milestones"];
        for i in range(len(steps)):
            steps[i] = steps[i]*num_batches

        
        system_dict["local"]["learning_rate_scheduler"] = mx.lr_scheduler.MultiFactorScheduler(
            step = steps,
            factor=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["gamma"],
            base_lr=learning_rate);

        #to make sure milestones value is restored to original one and not the updated steps while writing to experiment state
        #for i in range(len(steps)):
        #    steps[i] = steps[i]//num_batches

    return system_dict;


#Import Necessary Functions
from monk.system.imports import *
from monk.system.base_class import system
from monk.system.common import read_json
from monk.system.graphs.line import training_accuracy_curve
from monk.system.graphs.line import validation_accuracy_curve
from monk.system.graphs.line import training_loss_curve
from monk.system.graphs.line import validation_loss_curve
from monk.system.graphs.bar import training_time_plot
from monk.system.graphs.bar import max_accuracy_plot
from monk.system.graphs.bar import max_gpu_usage_plot




class compare(system):
    '''
    Class to compare multiple experiments

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose)
    

    @error_checks(None, ["name", ["A-Z", "a-z", "0-9", "-", "_"]], post_trace=False)
    @accepts("self", str, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Comparison(self, comparison_name):
        '''
        Create comparison project to compare and analyse multiple experiments 

        Args:
            comparison_name (str): Project Name

        Returns:
            None
        '''  
        self.set_system_comparison(comparison_name); 
        self.custom_print("Comparison: - {}".format(comparison_name));


    @accepts("self", str, str, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Add_Experiment(self, project_name, experiment_name):
        '''
        Add experiment for comparison

        Args:
            project_name (str): Project Name
            experiment_name (str): Experiment Name

        Returns:
            None
        '''  
        json_file = self.system_dict["master_systems_dir_relative"] + project_name + "/" + experiment_name + "/experiment_state.json";
        if(not os.path.isfile(json_file)):
            msg = "Project - {}, Experiment - {} does not exist".format(project_name, experiment_name)
            raise ConstraintError(msg)
        self.system_dict["local"]["experiments_list"].append(json_file);
        self.system_dict["local"]["project_experiment_list"].append(project_name + ":" + experiment_name);
        self.custom_print("Project - {}, Experiment - {} added".format(project_name, experiment_name));


    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def Generate_Statistics(self):
        '''
        Generate Statistics

        Args:
            None

        Returns:
            None
        ''' 
        self.custom_print("Generating statistics...");
        data = [];
        for i in range(len(self.system_dict["local"]["experiments_list"])):
            fname = self.system_dict["local"]["experiments_list"][i];
            system_dict = read_json(fname);
            data.append(system_dict);


        training_accuracy_curve(data, self.system_dict);
        validation_accuracy_curve(data, self.system_dict);
        training_loss_curve(data, self.system_dict);
        validation_loss_curve(data, self.system_dict);        
        
        training_time_plot(data, self.system_dict);
        max_accuracy_plot(data, self.system_dict);
        max_gpu_usage_plot(data, self.system_dict);
        
        
        # table
        table = [];

        #headers
        headers = ["project", "experiment", "base_model", "origin", "best val acc", "test acc", "num test images", \
        "epochs", "base lr", "optimizer", "lr scheduler", "loss func", "All layers trained", "gpu used", "max gpu usage", "training time", "train dataset type", \
        "num train images", "num val images", "shuffled dataset", "train transforms", "val transforms", "test transforms"] 


        for i in range(len(data)):
            tmp = [];
            tmp.append(str(data[i]["project_name"]));
            tmp.append(str(data[i]["experiment_name"]));
            model_name = data[i]["model"]["params"]["model_name"];
            if("/" in model_name):
                model_name = model_name.split("/")[-1];
            tmp.append(str(model_name));
            tmp.append(str(data[i]["origin"]));
            if(data[i]["training"]["status"]):
                tmp.append(str(data[i]["training"]["outputs"]["best_val_acc"]));
            else:
                tmp.append("NA");
            if(data[i]["testing"]["status"]):
                tmp.append(str(data[i]["testing"]["percentage_accuracy"]));
                tmp.append(str(data[i]["testing"]["num_images"]));
            else:
                tmp.append("NA");
                tmp.append("NA");
            tmp.append(str(data[i]["hyper-parameters"]["num_epochs"]));
            tmp.append(str(data[i]["hyper-parameters"]["optimizer"]["params"]["lr"]));
            tmp.append(str(data[i]["hyper-parameters"]["optimizer"]["name"]));
            tmp.append(str(data[i]["hyper-parameters"]["learning_rate_scheduler"]["name"]));
            tmp.append(str(data[i]["hyper-parameters"]["loss"]["name"]));
            tmp.append(str(data[i]["model"]["params"]["freeze_base_network"]));
            tmp.append(str(data[i]["model"]["params"]["use_gpu"]));
            if(data[i]["training"]["status"]):
                tmp.append(str(data[i]["training"]["outputs"]["training_time"]));
                tmp.append(str(data[i]["training"]["outputs"]["max_gpu_usage"]));
            else:
                tmp.append("NA");
                tmp.append("NA");
            tmp.append(str(data[i]["dataset"]["dataset_type"]));
            tmp.append(str(data[i]["dataset"]["params"]["num_train_images"]));
            tmp.append(str(data[i]["dataset"]["params"]["num_val_images"]));
            tmp.append(str(data[i]["dataset"]["params"]["train_shuffle"]));
            tmp.append(str(data[i]["dataset"]["transforms"]["train"]));
            tmp.append(str(data[i]["dataset"]["transforms"]["val"]));
            tmp.append(str(data[i]["dataset"]["transforms"]["test"]));
            table.append(tmp);

        my_df = pd.DataFrame(table);
        fname =  self.system_dict["master_comparison_dir_relative"] + "comparison.csv";
        my_df.to_csv(fname, index=False, header=headers);

        self.custom_print("Generated");
        self.custom_print("");


from monk.pytorch.losses.imports import *
from monk.system.imports import *


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def retrieve_loss(system_dict):
    '''
    Retrieve loss post state changes

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["criterion"] = system_dict["hyper-parameters"]["loss"]["name"];
    name = system_dict["local"]["criterion"];


    if(name == "l1"):
        system_dict["local"]["criterion"] = torch.nn.L1Loss(
            reduction='mean');

    elif(name == "l2"):
        system_dict["local"]["criterion"] = torch.nn.MSELoss(
            reduction='mean');

    elif(name == "softmaxcrossentropy"):
        system_dict["local"]["criterion"] = torch.nn.CrossEntropyLoss(
            weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"], 
            reduction='mean');

    elif(name == "crossentropy"):
        system_dict["local"]["criterion"] = torch.nn.NLLLoss(
            weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"], 
            reduction='mean');

    elif(name == "sigmoidbinarycrossentropy"):
        system_dict["local"]["criterion"] = torch.nn.BCEWithLogitsLoss(
            weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"], 
            reduction='mean');

    elif(name == "binarycrossentropy"):
        system_dict["local"]["criterion"] = torch.nn.BCELoss(
            weight=system_dict["hyper-parameters"]["loss"]["params"]["weight"], 
            reduction='mean');

    elif(name == "kldiv"):
        system_dict["local"]["criterion"] = torch.nn.KLDivLoss(
            reduction='mean');

    elif(name == "poissonnll"):
        system_dict["local"]["criterion"] = torch.nn.PoissonNLLLoss(
            log_input=system_dict["hyper-parameters"]["loss"]["params"]["log_pre_applied"],
            reduction='mean');

    elif(name == "huber"):
        system_dict["local"]["criterion"] = torch.nn.SmoothL1Loss(
            reduction='mean');

    elif(name == "hinge"):
        system_dict["local"]["criterion"] = torch.nn.HingeEmbeddingLoss(
            margin=system_dict["hyper-parameters"]["loss"]["params"]["margin"],
            reduction='mean');

    elif(name == "squaredhinge"):
        system_dict["local"]["criterion"] = torch.nn.SoftMarginLoss(
            reduction='mean');

    elif(name == "multimargin"):
        system_dict["local"]["criterion"] = torch.nn.MultiMarginLoss(
            p=1,
            margin=system_dict["hyper-parameters"]["loss"]["params"]["margin"],
            reduction='mean');

    elif(name == "squaredmultimargin"):
        system_dict["local"]["criterion"] = torch.nn.MultiMarginLoss(
            p=2,
            margin=system_dict["hyper-parameters"]["loss"]["params"]["margin"],
            reduction='mean');

    elif(name == "multilabelmargin"):
        system_dict["local"]["criterion"] = torch.nn.MultiLabelMarginLoss(
            reduction='mean');

    elif(name == "multilabelsoftmargin"):
        system_dict["local"]["criterion"] = torch.nn.MultiLabelSoftMarginLoss(
            reduction='mean');

    return system_dict;


from monk.tf_keras_1.training.imports import *
from monk.system.imports import *


class TimeHistory(krc.Callback):
    def __init__(self, log_dir=None):
        super().__init__()
        if(log_dir):
            self.log_file = log_dir + "times.txt";
            self.f = open(self.log_file, 'a');
        else:
            self.log_file=None
        

    def on_train_begin(self, logs={}):
        self.times = [];        

    def on_train_end(self, logs={}):
        if(self.log_file):
            self.f.close();        

    def on_epoch_begin(self, epoch, logs={}):
        self.epoch_time_start = time.time()

    def on_epoch_end(self, epoch, logs={}):
        self.times.append(time.time() - self.epoch_time_start)
        if(self.log_file):
            self.f.write(str(time.time() - self.epoch_time_start) + "\n");



class MemoryHistory(krc.Callback):
    def __init__(self):
        super().__init__()
        self.max_gpu_usage=0;
        
    def on_train_begin(self, logs={}):
        return
 
    def on_train_end(self, logs={}):
        return
 
    def on_epoch_begin(self, batch, logs={}):
        return
 
    def on_epoch_end(self, batch, logs={}):
        if(os.getcwd() == "/kaggle/working"):
            import GPUtil
            GPUs = GPUtil.getGPUs()
            if(len(GPUs) > 0):
                gpuMemoryUsed = GPUs[0].memoryUsed
                if(self.max_gpu_usage < int(gpuMemoryUsed)):
                    self.max_gpu_usage = int(gpuMemoryUsed);
        else:
            self.max_gpu_usage = 0;        
        return


from monk.tf_keras_1.schedulers.imports import *
from monk.system.imports import *
from monk.tf_keras_1.schedulers.common import StepDecay
from monk.tf_keras_1.schedulers.common import PolynomialDecay


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def load_scheduler(system_dict):
    '''
    Load schedulers for training state

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    learning_rate_scheduler = system_dict["local"]["learning_rate_scheduler"];
    optimizer = system_dict["local"]["optimizer"];
    learning_rate = system_dict["hyper-parameters"]["learning_rate"];

    if(learning_rate_scheduler == "steplr"):
        system_dict["local"]["learning_rate_scheduler"] = StepDecay(
                initAlpha=learning_rate, 
                factor=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["gamma"], 
                dropEvery=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["step_size"]);

    elif(learning_rate_scheduler == "exponentiallr"):
        system_dict["local"]["learning_rate_scheduler"] = PolynomialDecay(
                maxEpochs=100, 
                initAlpha=learning_rate, 
                power=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["gamma"]);

    elif(learning_rate_scheduler == "reduceonplateaulr"):
        system_dict["local"]["learning_rate_scheduler"] = krc.ReduceLROnPlateau(
                monitor='val_loss', 
                factor=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["factor"], 
                patience=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["patience"], 
                verbose=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["verbose"], 
                mode=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["mode"], 
                min_delta=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["threshold"], 
                cooldown=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["cooldown"], 
                min_lr=system_dict["hyper-parameters"]["learning_rate_scheduler"]["params"]["min_lr"]);


    return system_dict;





import os 
import sys 
import numpy as np 

import warnings
warnings.filterwarnings("ignore", category=FutureWarning)
warnings.simplefilter(action='ignore', category=Warning)

from torchvision import transforms


